@@ -356,15 +356,29 @@ fn_get_arguments(ProxyFunction *func,
 
 	for (i = 0; i < total; i++)
 	{
-		if (modes && modes[i] == 'o')
-			continue;
-		type = plproxy_find_type_info(func, types[i], 1);
-		pos = func->arg_count++;
-		func->arg_types[pos] = type;
-		if (names && names[i])
-			func->arg_names[pos] = plproxy_func_strdup(func, names[i]);
-		else
-			func->arg_names[pos] = NULL;
+		char mode = modes ? modes[i] : PROARGMODE_IN;
+		switch (mode) {
+		case PROARGMODE_IN:
+		case PROARGMODE_INOUT:
+			type = plproxy_find_type_info(func, types[i], 1);
+			pos = func->arg_count++;
+			func->arg_types[pos] = type;
+			if (names && names[i])
+				func->arg_names[pos] = plproxy_func_strdup(func, names[i]);
+			else
+				func->arg_names[pos] = NULL;
+			break;
+		case PROARGMODE_VARIADIC:
+			elog(ERROR, "PL/Proxy does not support variadic args");
+			break;
+		case PROARGMODE_OUT:
+		case PROARGMODE_TABLE:
+			/* output args, ignore */
+			break;
+		default:
+			elog(ERROR, "PL/Proxy: unknown value in proargmodes: %c", mode);
+			break;
+		}
 	}
 }
 
@@ -426,13 +440,14 @@ fn_refresh_record(FunctionCallInfo fcinfo,
 	TypeFuncClass rtc;
 	TupleDesc tuple_current, tuple_cached;
 	MemoryContext old_ctx;
+	Oid tuple_oid;
 	int natts;
 
 	/*
 	 * Compare cached tuple to current one.
 	 */
 	tuple_cached = func->ret_composite->tupdesc;
-	rtc = get_call_result_type(fcinfo, NULL, &tuple_current);
+	rtc = get_call_result_type(fcinfo, &tuple_oid, &tuple_current);
 	Assert(rtc == TYPEFUNC_COMPOSITE);
 	if (equalTupleDescs(tuple_current, tuple_cached))
 		return;
@@ -448,7 +463,7 @@ fn_refresh_record(FunctionCallInfo fcinfo,
 	pfree(func->remote_sql);
 
 	/* construct new data */
-	func->ret_composite = plproxy_composite_info(func, tuple_current);	
+	func->ret_composite = plproxy_composite_info(func, tuple_current);
 	natts = func->ret_composite->tupdesc->natts;
 	func->result_map = plproxy_func_alloc(func, natts * sizeof(int));
 	func->remote_sql = plproxy_standard_query(func, true);
@@ -464,7 +479,7 @@ fn_compile(FunctionCallInfo fcinfo,
 	Form_pg_proc proc_struct;
 
 	proc_struct = (Form_pg_proc) GETSTRUCT(proc_tuple);
-	if (proc_struct->provolatile != 'v')
+	if (proc_struct->provolatile != PROVOLATILE_VOLATILE)
 		elog(ERROR, "PL/Proxy functions must be volatile");
 
 	f = fn_new(fcinfo, proc_tuple);
@@ -551,6 +566,11 @@ plproxy_compile(FunctionCallInfo fcinfo, bool validate)
 		/* in case of untyped RECORD, check if cached type is valid */
 		fn_refresh_record(fcinfo, f, proc_tuple);
 	}
+	else if (f->ret_composite)
+	{
+		if (!plproxy_composite_valid(f->ret_composite))
+			fn_refresh_record(fcinfo, f, proc_tuple);
+	}
 
 	ReleaseSysCache(proc_tuple);
 
