@@ -37,12 +37,17 @@
 #include <catalog/pg_user_mapping.h>
 #endif
 
+#if PG_VERSION_NUM >= 90300
+#include <access/htup_details.h>
+#endif
+
 #include <access/reloptions.h>
 #include <access/tupdesc.h>
 #include <catalog/pg_namespace.h>
 #include <catalog/pg_proc.h>
 #include <catalog/pg_type.h>
 #include <commands/trigger.h>
+#include <lib/stringinfo.h>
 #include <mb/pg_wchar.h>
 #include <miscadmin.h>
 #include <nodes/value.h>
@@ -95,6 +100,44 @@
 #endif
 #endif
 
+/*
+ * backwards compat with 8.4
+ */
+#ifndef PROARGMODE_IN
+#define PROARGMODE_IN       'i'
+#endif
+#ifndef PROARGMODE_OUT
+#define PROARGMODE_OUT      'o'
+#endif
+#ifndef PROARGMODE_INOUT
+#define PROARGMODE_INOUT    'b'
+#endif
+#ifndef PROARGMODE_VARIADIC
+#define PROARGMODE_VARIADIC 'v'
+#endif
+#ifndef PROARGMODE_TABLE
+#define PROARGMODE_TABLE    't'
+#endif
+
+#ifndef TYPTYPE_BASE
+#define TYPTYPE_BASE 'b'
+#endif
+#ifndef TYPTYPE_COMPOSITE
+#define TYPTYPE_COMPOSITE 'c'
+#endif
+#ifndef TYPTYPE_DOMAIN
+#define TYPTYPE_DOMAIN 'd'
+#endif
+#ifndef TYPTYPE_ENUM
+#define TYPTYPE_ENUM 'e'
+#endif
+#ifndef TYPTYPE_PSEUDO
+#define TYPTYPE_PSEUDO 'p'
+#endif
+#ifndef TYPTYPE_RANGE
+#define TYPTYPE_RANGE 'r'
+#endif
+
 /*
  * Determine if this argument is to SPLIT
  */
@@ -169,6 +212,7 @@ typedef struct ProxyConnectionState {
 	time_t		query_time;		/* When last query was sent */
 	bool		same_ver;		/* True if dest backend has same X.Y ver */
 	bool		tuning;			/* True if tuning query is running on conn */
+	bool		waitCancel;		/* True if waiting for answer from cancel */
 } ProxyConnectionState;
 
 /* Single database connection */
@@ -296,7 +340,10 @@ typedef struct ProxyComposite
 	TupleDesc	tupdesc;		/* Return tuple descriptor */
 	ProxyType **type_list;		/* Column type info */
 	char	  **name_list;		/* Quoted column names */
+	int			nfields;		/* number of non-dropped fields */
 	bool		use_binary;		/* True if all columns support binary recv */
+	bool		alterable;		/* if it's real table that can change */
+	RowStamp	stamp;
 } ProxyComposite;
 
 /* Temp structure for query parsing */
@@ -426,6 +473,7 @@ Datum		plproxy_recv_type(ProxyType *type, char *str, int len, bool bin);
 HeapTuple	plproxy_recv_composite(ProxyComposite *meta, char **values, int *lengths, int *fmts);
 void		plproxy_free_type(ProxyType *type);
 void		plproxy_free_composite(ProxyComposite *meta);
+bool		plproxy_composite_valid(ProxyComposite *type);
 
 /* cluster.c */
 void		plproxy_cluster_cache_init(void);
