@@ -183,6 +183,8 @@ plproxy_standard_query(ProxyFunction *func, bool add_types)
 		ProxyComposite *t = func->ret_composite;
 		for (i = 0; i < t->tupdesc->natts; i++)
 		{
+			if (t->tupdesc->attrs[i]->attisdropped)
+				continue;
 			appendStringInfo(&sql, "%s%s::%s",
 							 ((i > 0) ? ", " : ""),
 							 t->name_list[i],
@@ -217,6 +219,8 @@ plproxy_standard_query(ProxyFunction *func, bool add_types)
 		appendStringInfo(&sql, " as (");
 		for (i = 0; i < t->tupdesc->natts; i++)
 		{
+			if (t->tupdesc->attrs[i]->attisdropped)
+				continue;
 			appendStringInfo(&sql, "%s%s %s",
 							((i > 0) ? ", " : ""),
 							t->name_list[i],
